import curses
import json
import os
from datetime import datetime
from getpass import getpass
from pathlib import Path
from typing import Optional, List, Tuple

from youtyper.analyzers import default_analyzers
from youtyper.lessons import default_lesson_generators
from youtyper.lessons.lessons import TextLessonGenerator, LessonGenerator, Lesson
from .ui import UI

import click


@click.command(context_settings=dict(ignore_unknown_options=True))
@click.option(
    "--lesson_type",
    "-t",
    default="built-in",
    type=click.Choice(["built-in", "text", "python"], case_sensitive=False),
    help="How the lesson is generated",
)
@click.option(
    "--lesson_name",
    "-ln",
    default="default_lesson",
    type=click.Choice(default_lesson_generators.keys(), case_sensitive=True),
    help="Choose built-in lesson to use",
)
@click.option(
    "--text_path",
    "-t",
    default="",
    type=str,
    help='Text file path to load into lessons.  Available if lesson_type="text"',
)
@click.option(
    "--disable_shuffle",
    "-ds",
    is_flag=True,
    help='Disable shuffling of lessons. Available if lesson_type="text"',
)
@click.option(
    "--num_lessons",
    "-n",
    default=0,
    type=int,
    help='Number of lessons. default: "0": never quit until all lessons are finished. Available if lesson_type="text"',
)
@click.option(
    "--len_lessons",
    "-l",
    default="100",
    type=int,
    help='Number of maximum characters of each lesson. Available if lesson_type="text"',
)
@click.option(
    "--analyzer",
    "-a",
    default=["cpm", "error_rate"],
    type=click.Choice(default_analyzers.keys(), case_sensitive=True),
    multiple=True,
    help="Analytics to be shown at the end of the lesson",
)
@click.option(
    "--ignore_consecutive_errors",
    "-i",
    is_flag=True,
    help="Ignore consecutive errors on calculating analytics",
)
@click.option(
    "--custom_analyzer",
    "-ca",
    type=(str, str),
    multiple=True,
    help="Custom analyzer path and class name",
)
@click.option(
    "--custom_lesson_generator",
    "-cl",
    type=(str, str),
    default=(None, None),
    help='Custom lessson generator path and class name. Available if lesson_type="python"',
)
@click.argument("custom_args", nargs=-1, type=click.UNPROCESSED)
def main(
    lesson_type: str,
    lesson_name: str,
    text_path: str,
    disable_shuffle: bool,
    num_lessons: Optional[int],
    len_lessons: int,
    analyzer: List[str],
    ignore_consecutive_errors: bool,
    custom_analyzer: List[Tuple[str, str]],
    custom_lesson_generator: Tuple[str, str],
    custom_args,
):
    if custom_lesson_generator[0] is None and custom_lesson_generator[1] is None:
        custom_lesson_generator = []
    # Create LessonGenerator
    if lesson_type == "built-in":
        lesson_generator = default_lesson_generators[lesson_name]()
    elif lesson_type == "text":
        if not text_path:
            raise ValueError("You must specify text_path to generate lessons.")
        if not os.path.exists(text_path):
            raise FileNotFoundError("You must specify valid text_path.")
        try:
            text = open(text_path, encoding="ascii").read()
        except UnicodeDecodeError as e:
            print(e.__traceback__)
            raise ValueError("Only accepts ascii file for text_path")
        filename = os.path.splitext(text_path)[0]
        lesson_generator = TextLessonGenerator(
            text, filename, num_lessons, len_lessons, not disable_shuffle
        )
    elif lesson_type == "python":
        f = open(custom_lesson_generator[0])
        code = f.read()
        exec(code, globals())
        custom_lesson_generator_class = globals()[custom_lesson_generator[1]]
        lesson_generator: LessonGenerator = custom_lesson_generator_class(custom_args)
    else:
        raise

    # Create Analyzer
    analyzers = [default_analyzers[name]() for name in analyzer]

    for analyzer_path, analyzer_classname in custom_analyzer:
        f = open(analyzer_path[0])
        code = f.read()
        exec(code, globals())
        custom_analyzer_class = globals()[analyzer_classname]
        custom_analyzer = custom_analyzer_class(custom_args)
        analyzers.append(custom_analyzer)

    try:
        p = UI()
        logs = []
        current_lesson = 1
        win = curses.initscr()
        curses.noecho()
        win.clear()
        win.refresh()
        win.clear()
        win.refresh()
        message = (
            f"press enter to start lesson {current_lesson} / {len(lesson_generator)}"
        )
        win.addstr(0, 0, message)
        key = win.getkey()
        while l := next(lesson_generator):
            l = l  # type:Lesson
            start_time = datetime.now()
            lesson_log = p.start(l)
            if lesson_log is None:
                print("\nAborted!")
                raise SystemExit()
            data = {}
            lines = 0
            win.clear()
            win.refresh()
            message = f"lesson {current_lesson} / {len(lesson_generator)} result:"
            win.addstr(lines, 0, message)
            for a in analyzers:
                json_result, printable_result = a.analyze(lesson_log=lesson_log)
                win.addstr(lines + 1, 0, printable_result)
                lines += len(printable_result.split("\n"))
                data[a.get_analytics_name()] = json_result
            log = {
                "lesson_name": l.lesson_name,
                "lesson_id": l.lesson_id,
                "command-line-options": {
                    "lesson_type": lesson_type,
                    "lesson_name": lesson_name,
                    "text_path": text_path,
                    "disable_shuffle": disable_shuffle,
                    "num_lessons": num_lessons,
                    "len_lessons": len_lessons,
                    "analyzer": analyzer,
                    "ignore_consecutive_errors": ignore_consecutive_errors,
                    "custom_analyzer": custom_analyzer,
                    "custom_lesson_generator": custom_lesson_generator,
                    "custom_args": custom_args,
                },
                "text": l.text,
                "events": [e.to_dict() for e in lesson_log.events],
                "analytics": data,
            }
            logs.append(log)
            abs_dir = str(Path.home()) + "/.youtyper/"
            os.makedirs(abs_dir, exist_ok=True)
            with open(
                f"{abs_dir}/{start_time:%Y%m%d_%H%M%S}_{l.lesson_name}_{l.lesson_id}.json",
                "w",
            ) as f:
                f.write(json.dumps(log))
            current_lesson += 1
            message = f"press enter to start lesson {current_lesson} / {len(lesson_generator)}"
            win.addstr(lines + 1, 0, message)
            key = win.getkey()
    finally:
        curses.endwin()


if __name__ == "__main__":
    main()



from abc import ABCMeta, abstractmethod
from typing import List, Dict

from youtyper.logs import LessonLog


class Analyzer(metaclass=ABCMeta):
    @abstractmethod
    def __init__(self):
        pass

    @abstractmethod
    def analyze(self, lesson_log: LessonLog) -> (Dict, str):
        pass

    @staticmethod
    @abstractmethod
    def get_analytics_name() -> str:
        return ""



import curses, time
from typing import Optional

from .lessons.lessons import Lesson
from .logs import LessonLog


class UI(object):
    def __init__(self) -> None:
        pass

    def start(self, lesson: Lesson) -> Optional[LessonLog]:
        text = lesson.text
        lesson_log = LessonLog()
        current_str = ""
        aborted = False
        try:
            win = curses.initscr()
            curses.noecho()
            missed = False
            key = ""
            target = ""
            while True:
                win.clear()
                if missed:
                    win.addstr(0, 0, "miss! " + key + "->" + target)
                win.addstr(1, 0, "Lesson: ")
                win.addstr(2, 0, "You:    ")
                win.addstr(1, 8, text)
                win.addstr(2, 8, current_str)
                time.sleep(0.05)
                key = win.getkey()
                target = text[len(current_str)]
                lesson_log.record_key(key, target)
                win.refresh()
                if key == target:
                    missed = False
                    current_str = current_str + key
                elif key:
                    missed = True
                else:
                    pass
                if current_str == text:
                    break
        except:
            aborted = True
        finally:
            curses.endwin()
            return lesson_log if not aborted else None



from abc import abstractmethod
from random import choices, sample
from typing import Optional

from youtyper.assets.famous_texts import (
    universal_declaration_of_human_rights_text,
    pangrams,
)
from youtyper.lessons.lessons import LessonGenerator, Lesson, TextLessonGenerator


class PangramLessonGenerator(LessonGenerator):
    def __init__(self):
        super().__init__()
        self.n = len(pangrams)
        self.c = 0
        self.pangrams = sample(pangrams, len(pangrams))

    def __next__(self) -> Optional[Lesson]:
        if self.c >= self.n:
            return None
        self.c += 1
        return Lesson(
            text=pangrams[self.c],
            lesson_id=self.get_generator_name() + "_" + str(self.c),
            lesson_name=self.get_generator_name(),
        )

    def __len__(self):
        return self.n

    @staticmethod
    def get_generator_name() -> str:
        return "default_lesson"


class FamousTextsLessonGenerator(TextLessonGenerator):
    def __init__(self):
        super().__init__(self.get_generator_name(), self.get_text(), _shuffle=False)

    def __next__(self) -> Optional[Lesson]:
        if self.c >= self.n:
            return None
        self.c += 1
        return Lesson(
            text=self.text[self.c],
            lesson_id=self.get_generator_name() + "_" + str(self.c),
            lesson_name=self.get_generator_name(),
        )

    def __len__(self):
        return self.n

    @staticmethod
    @abstractmethod
    def get_generator_name() -> str:
        return ""

    @staticmethod
    @abstractmethod
    def get_text() -> str:
        return ""


class UniversalDeclarationOfHumanRightsLessonGenerator(FamousTextsLessonGenerator):
    def __init__(self):
        super().__init__()

    @staticmethod
    def get_generator_name() -> str:
        return "Universal_Declaration_Of_Human_Rights"

    @staticmethod
    def get_text() -> str:
        return universal_declaration_of_human_rights_text



from youtyper.lessons.basic import (
    PangramLessonGenerator,
    UniversalDeclarationOfHumanRightsLessonGenerator,
)

default_lesson_generator_classes = [
    PangramLessonGenerator,
    UniversalDeclarationOfHumanRightsLessonGenerator,
]

default_lesson_generators = {
    lg.get_generator_name(): lg for lg in default_lesson_generator_classes
}



from abc import ABCMeta, abstractmethod
from typing import Optional
from random import shuffle


class Lesson(object):
    def __init__(self, text: str, lesson_id: str, lesson_name: str) -> None:
        self.text = text
        self.lesson_id = lesson_id
        self.lesson_name = lesson_name


class LessonGenerator(metaclass=ABCMeta):
    @abstractmethod
    def __init__(self):
        pass

    @abstractmethod
    def __next__(self) -> Optional[Lesson]:
        pass

    @abstractmethod
    def __len__(self) -> int:
        pass

    def __iter__(self):
        return self

    @staticmethod
    @abstractmethod
    def get_generator_name() -> str:
        return ""


class TextLessonGenerator(LessonGenerator):
    def __init__(
        self,
        lesson_name: str,
        text: str,
        num_lessons: int = 0,
        len_lessons: int = 100,
        _shuffle: bool = True,
    ):
        super().__init__()
        self.lesson_name = lesson_name
        processed_texts = []
        for line in text.split("\n"):
            while len(line) >= len_lessons:
                if l := line[:len_lessons]:
                    processed_texts.append(l)
                line = line[len_lessons:]
            if line:
                processed_texts.append(line)
        self.text = processed_texts
        if _shuffle:
            shuffle(self.text)
        if not num_lessons:
            self.n = len(self.text)
        else:
            self.n = min(len(self.text), num_lessons)
        self.c = 0

    def __next__(self) -> Optional[Lesson]:
        if self.n and self.c >= self.n:
            return None
        self.c += 1
        return Lesson(
            text=self.text[self.c],
            lesson_id=self.lesson_name + "_" + str(self.c),
            lesson_name=self.lesson_name,
        )

    def __len__(self):
        return self.n

    @staticmethod
    def get_generator_name() -> str:
        return "text"



from typing import Dict

from youtyper.analyzers.basic import (
    CharacterPerMinuteAnalyzer,
    BuiltInAnalyzer,
    ErrorRateAnalyzer,
)

default_analyzer_classes = [CharacterPerMinuteAnalyzer, ErrorRateAnalyzer]

default_analyzers: Dict[str, BuiltInAnalyzer] = {
    a.get_abbreviated_name(): a for a in default_analyzer_classes
}



from datetime import datetime
from typing import List


class KeyStrokeLog(object):
    def __init__(self, timestamp: datetime, key: str, target: str):
        self.timestamp = timestamp
        self.key = key
        self.target = target

    def to_dict(self):
        return {
            "timestamp": str(self.timestamp),
            "key": self.key,
            "targert": self.target,
        }


class LessonLog(object):
    def __init__(self) -> None:
        self.events: List[KeyStrokeLog] = []

    def record_key(self, key: str, target: str, timestamp: datetime = None):
        if timestamp is None:
            timestamp = datetime.now()
        self.events.append(KeyStrokeLog(timestamp, key, target))



import setuptools

PACKAGE_NAME = "youtyper"

setuptools.setup(
    name=PACKAGE_NAME,
    version="0.0.3",
    description="A customizable command line touch typing tutor",
    long_description=open("README.md").read().strip(),
    long_description_content_type="text/markdown",
    author="hmirin",
    url="http://github.com/hmirin/youtyper",
    python_requires=">=3.8",
    py_modules=[PACKAGE_NAME],
    install_requires=["click>=7"],
    license="MIT License",
    zip_safe=False,
    keywords="touch-typing typing",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Environment :: Console :: Curses",
        "Intended Audience :: Developers",
        "Intended Audience :: Education",
        "Intended Audience :: End Users/Desktop",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3.8",
        "Topic :: Utilities",
        "Topic :: Education",
    ],
    packages=setuptools.find_packages(exclude=["contrib", "docs", "tests*"]),
    entry_points="""
                    [console_scripts]
                    {app}={pkg}.main:main
                 """.format(
        app=PACKAGE_NAME, pkg=PACKAGE_NAME
    ),
)



from abc import abstractmethod
from collections import defaultdict
from datetime import timedelta
from typing import List, Dict, Optional

from youtyper.analyzers.analyzer import Analyzer

from youtyper.logs import KeyStrokeLog, LessonLog


class BuiltInAnalyzer(Analyzer):
    @staticmethod
    @abstractmethod
    def get_abbreviated_name() -> str:
        return ""


def time_to_push_correct_key(events: List[KeyStrokeLog]):
    last_time = None
    accumulated_time = timedelta()
    elasped_time_dict: Dict[str, List[timedelta]] = defaultdict(list)
    for key_stroke_log in events:
        if last_time is None:
            last_time = key_stroke_log.timestamp
        else:
            current_time = key_stroke_log.timestamp
            elasped_time = current_time - last_time
            if key_stroke_log.target == key_stroke_log.key:
                elasped_time_dict[key_stroke_log.target].append(
                    elasped_time + accumulated_time
                )
                accumulated_time = timedelta()
            else:
                accumulated_time += elasped_time
    return elasped_time_dict


def elasped_seconds_and_num_keys(
    lesson_log: LessonLog, key: str = None, tail: Optional[int] = None
) -> (int, timedelta):
    if tail is not None:
        events = lesson_log.events[-tail:]
    else:
        events = lesson_log.events
    if key is None:
        first_stroke_time = events[0].timestamp
        last_stroke_time = events[-1].timestamp
        return (len(events) - 1), last_stroke_time - first_stroke_time
    else:
        seconds = [t.total_seconds() for t in time_to_push_correct_key(events)[key]]
        return len(seconds), timedelta(seconds=sum(seconds))


class CharacterPerMinuteAnalyzer(BuiltInAnalyzer):
    def __init__(self):
        super().__init__()

    def analyze(self, lesson_log: LessonLog) -> (Dict, str):
        num_keys, elasped_time = elasped_seconds_and_num_keys(lesson_log)
        seconds = elasped_time.total_seconds()
        cpm = num_keys / (seconds / 60)
        return (
            {"cpm": cpm, "len_lesson": num_keys, "total_time_in_seconds": seconds},
            f"cpm: {cpm:.1f} (elasped seconds: {elasped_time}, len_lesson: {num_keys})",
        )

    @staticmethod
    def get_analytics_name() -> str:
        return "Character Per Minute (CPM)"

    @staticmethod
    def get_abbreviated_name() -> str:
        return "cpm"


class ErrorRateAnalyzer(BuiltInAnalyzer):
    def __init__(self):
        super().__init__()

    def analyze(self, lesson_log: LessonLog) -> (Dict, str):
        events = lesson_log.events
        wrong_events = [e for e in events if e.target != e.key]
        error_rate = len(wrong_events) / len(events)
        return (
            {
                "error_rate": error_rate,
                "total_push": len(events),
                "wrong_push": len(wrong_events),
            },
            f"error_rate: {error_rate*100:.1f}% (wrong push: {len(wrong_events)}, total push: {len(events)})",
        )

    @staticmethod
    def get_analytics_name() -> str:
        return "Error Rate"

    @staticmethod
    def get_abbreviated_name() -> str:
        return "error_rate"
